"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const constants_1 = require("@scrabble-solver/constants");
const Cell_1 = __importDefault(require("./Cell"));
const Tile_1 = __importDefault(require("./Tile"));
class Board {
    static create(size) {
        return Board.fromStringArray(Array(size).fill(Array(size).fill(' ').join('')));
    }
    static fromJson(json) {
        return new Board({
            rows: json.map((row) => row.map(Cell_1.default.fromJson)),
        });
    }
    static fromStringArray(stringArray) {
        return new Board({
            rows: stringArray.map((row, y) => row.split('').map((character, x) => new Cell_1.default({
                isEmpty: !character || character === constants_1.EMPTY_CELL,
                tile: character === constants_1.EMPTY_CELL ? Tile_1.default.Null : new Tile_1.default({ character }),
                x,
                y,
            }))),
        });
    }
    constructor({ rows }) {
        this.rows = rows;
        this.columnsCount = rows[0].length;
        this.rowsCount = rows.length;
    }
    get center() {
        const x = Math.floor(this.columnsCount / 2);
        const y = Math.floor(this.rowsCount / 2);
        return this.rows[y][x];
    }
    clone() {
        const rows = this.rows.map((row) => row.map((cell) => cell.clone()));
        return new Board({ rows });
    }
    collides(cell) {
        return this.collidesUp(cell) || this.collidesDown(cell) || this.collidesLeft(cell) || this.collidesRight(cell);
    }
    collidesDown({ x, y }) {
        return y < this.rowsCount - 1 && !this.rows[y + 1][x].isEmpty;
    }
    collidesLeft({ x, y }) {
        return x > 0 && !this.rows[y][x - 1].isEmpty;
    }
    collidesRight({ x, y }) {
        return x < this.columnsCount - 1 && !this.rows[y][x + 1].isEmpty;
    }
    collidesUp({ x, y }) {
        return y > 0 && !this.rows[y - 1][x].isEmpty;
    }
    equals(other) {
        return (this.columnsCount === other.columnsCount &&
            this.rowsCount === other.rowsCount &&
            this.rows.every((row, rowIndex) => {
                return row.every((cell, cellIndex) => {
                    return cell.equals(other.rows[rowIndex][cellIndex]);
                });
            }));
    }
    getBlanksCount() {
        return this.rows.reduce((count, row) => {
            return count + row.reduce((rowCount, cell) => rowCount + (cell.tile.isBlank ? 1 : 0), 0);
        }, 0);
    }
    getColumn(index) {
        return this.rows.map((row) => row[index]);
    }
    getRow(index) {
        return this.rows[index];
    }
    getTilesCount() {
        return this.rows.reduce((count, row) => {
            return count + row.reduce((rowCount, cell) => rowCount + (cell.hasTile() ? 1 : 0), 0);
        }, 0);
    }
    getWords() {
        const columns = [];
        for (let x = 0; x < this.columnsCount; ++x) {
            const column = [];
            for (let y = 0; y < this.rowsCount; ++y) {
                column.push(this.rows[y][x]);
            }
            columns.push(column);
        }
        const columnsBoard = new Board({ rows: columns });
        const lines = this.toString().split('\n').concat(columnsBoard.toString().split('\n'));
        const words = lines
            .flatMap((line) => line.replaceAll(/\s+/g, constants_1.EMPTY_CELL).split(' '))
            .filter((word) => word.length > 1);
        return words;
    }
    isEmpty() {
        return this.rows.every((row) => row.every(({ isEmpty }) => isEmpty));
    }
    toJson() {
        return this.rows.map((row) => row.map((cell) => cell.toJson()));
    }
    toString() {
        return this.rows.map((row) => row.map(String).join('')).join('\n');
    }
    updateCell(x, y, updateCell) {
        this.rows[y][x] = updateCell(this.rows[y][x]);
    }
    updateRow(y, updateRow) {
        this.rows[y] = updateRow(this.rows[y]);
    }
}
exports.default = Board;
